<?php
/**
 * jDiction library entry point
 *
 * @package jDiction
 * @link http://joomla.itronic.at
 * @copyright	Copyright (C) 2011 ITronic Harald Leithner. All rights reserved.
 * @license GNU General Public License v3
 * @version 1.6.1 (18)
 *
 * This file is part of jDiction.
 *
 * jDiction is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * jDiction is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDiction.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
defined('_JEXEC') or die;

define('JDICTION_TRANSLATION_STATUS_FULL', 1);
define('JDICTION_TRANSLATION_STATUS_OLD', 2);
define('JDICTION_TRANSLATION_STATUS_NONE', 3);
/* unused atm */
define('JDICTION_TRANSLATION_STATUS_PARTLY', 3);
define('JDICTION_TRANSLATION_STATUS_PUBLISHED', 5);
define('JDICTION_TRANSLATION_STATUS_UNPUBLISHED', 6);
define('JDICTION_TRANSLATION_STATUS_UNSET', 7);

//check for XML Support
if (function_exists("simplexml_load_string") && class_exists("SimpleXMLElement")) {
	define('JDICTION_TRANSLATION_USEXML', true);
} else {
	define('JDICTION_TRANSLATION_USEXML', false);
}

jimport('joomla.language.helper');
jimport('jdiction.helpers.translation');

class jDiction extends JObject {
	
	/** @var holds all tables which could be translated jDiction */
	protected $tables = array();
	
	/** @var holds all component task/view componation we can translate */
	protected $components = Array();
	
	/** @var array holding the cache for translations */
	protected $cache = Array();
	
	/** @var object local JDatabase object */
	protected $db;
	
	/**
	 * Dispatcher object
	 *
	 * @var object
	 */
	protected $dispatcher;
	
	/**
	 * defines if jDiction is ready
	 * @var bool
	 */
	protected $ready=false;
	
	/**
	 * traceLog for debuging
	 */
	protected $traceLog=array();
	
	/**
	 * current trace Id
	 */
	protected $traceId=false;

	/**
	 * Returns the global jDiction object, only creating it
	 * if it doesn't already exist.
	 *
	 * @return	object	The global jDiction object
	 */
	public static function getInstance() {
		static $instance;

		if (!isset ($instance)) {
			$instance = new jDiction();
		}
		
		return $instance;
	}
	
	/**
	 * Initialise jDiction
	 */
	public function initialise() {
		$this->WorkaroundDatabaseConfig();
		//load Lanuage file
		$lang = JFactory::getLanguage();
		$lang->load('lib_jdiction', JPATH_LIBRARIES.'/jdiction');
		
		// load dispatcher and jdiction plugins
		JPluginHelper::importPlugin('jdiction');
		$this->dispatcher = JDispatcher::getInstance();
		
		$this->dispatcher->trigger('onjDictionAfterInitialise');
		
		$this->loadTables();
		
		$this->ready = true;
	}
	
	public function WorkaroundDatabaseConfig() {
		if ($_REQUEST['option'] == 'com_config') {
			jimport('jdiction.form.fields.databaseconnection');
		}

	}
	
	/**
	 * check the status of jdiction
	 * @return bool the status of the library
	 */
	public function getStatus() {
		return $this->ready;
	}
	
	/** 
	 * load all core table definitions
	 */
	public function loadTables() {
		jimport('joomla.filesystem.folder');
		$dir = dirname(__FILE__).'/tables';

		//$subdirs = JFolder::files($dir,'.xml', false, true);
		//Workaround for Suhosin problem: ALERT - canary mismatch on efree() - heap overflow detected
		$subdir = array();
		$subdirstmp = scandir($dir);
		foreach($subdirstmp as $subdir) {
			if (substr($subdir, -4) == '.xml') {
				$subdirs[] = $dir.'/'.$subdir;
			}
		}
		//Workaround end
		
		$this->dispatcher->trigger('onjDictionBeforeLoadTables', array( &$subdirs));
		
		foreach ($subdirs as $path) {
			$this->parseTable($path);
		}

		$this->dispatcher->trigger('onjDictionAfterLoadTables', array( &$subdirs));

	}
	
	/**
	 * load xml definiation of a table with its component and views
	 */
	public function parseTable($xmlfile) {
		// Try to load the file
		if (!$xml = JFactory::getXML($xmlfile)) {
			return false;
		}

		// Check that it's a metadata file
		if ((string)$xml->getName() != 'jdiction') {
			return false;
		}
		
		$this->dispatcher->trigger('onjDictionParseTable', array( &$xml));

		$metadata = array();
		foreach ($xml->metadata->children() as $child) {
			$metadata[$child->getName()] = (string) $child;
		}
		//we don't need the prefix
		if (strpos($metadata['table'], '#__') === 0) {
			$metadata['table'] = substr($metadata['table'],3);
		}

		$this->tables[$metadata['table']] = $metadata['key'];
		$this->components[$metadata['component']]['__metadata'] = $metadata;
		
		//prepare component table
		foreach($xml->views[0] as $view) {
			foreach($view as $layout) {
				if ($layout['active'] == "false") {
					$this->components[$metadata['component']][(string)$view['name']][(string)$layout['name']] = false;
				} else {
					$this->components[$metadata['component']][(string)$view['name']][(string)$layout['name']] = $layout;
					if ($view['list']) {
						$this->components[$metadata['component']]['__metadata']['lists'][(string)$view['name']] = (string)$view['list'];
					}
				}
			}
		}
	}
	
	public function getComponent($component=null) {
		if (!$component) {
			return $this->components;
		} else {
			return $this->components[$component];
		}
	}
	
	/**
	 * Debug function to start and stop trace
	 */
	public function trace($start=true, $id=1) {
		$trace = true;

		if ($start) {
			$this->traceId = $id;
			if (array_key_exists($id, $this->traceLog)) {
				$this->traceLog[$id] = array();
			} else {
				$trace = $this->traceLog[$id];
			}
		} else {
			$trace = $this->traceLog[$id];
			unset($this->trace[$id]);
			$this->traceId = false;
		}
		return $trace;
	}
	
	public function addLog($msg) {
		if ($this->traceId) {
			$this->traceLog[$this->traceId] = array(
				'message' => $msg
			);
		}
	}
	
	/**
	 * translate content
	 */
	public function translate(&$rows, $language, &$metadata, $resulttype='BOTH', $key=null) {
		
		if (!$this->ready) {
			return false;
		}
		
		/* If we have no result we exit here */
		if (empty($rows)) {
			return false;
		}

		//find primary key for each table alias
		foreach($metadata['tables'] as $tablealias=>$table) {
			foreach($table['fields'] as $fieldid=>$field) {
				if (isset($field->orgname)) {
					$fieldName = $field->orgname;
				} else {
					$fieldName = $field->name;
				}
				
				if ($fieldName == $this->tables[$metadata['tablealias'][$tablealias]]) {
					$aliaskey[$tablealias] = $fieldid;
				}
			}
		}
		
		if (is_numeric($key) && $key < 0) {
			$rows = array($rows);
		}

		$newrows = array();
		//translate all rows
		if (is_array($rows) && count($rows)) {
			foreach($rows as $rowkey=>&$row) {
				$targetkey = $rowkey;
				foreach($metadata['tables'] as $tablealias=>$table) {
					if (array_key_exists($metadata['tablealias'][$tablealias], $this->tables)) {
						$translation = $this->getTranslation($metadata['tablealias'][$tablealias], $row[$aliaskey[$tablealias]], $language);
					} else {
						$translation = false;
					}
					/* Unset this row if it hast state -3 (hide) */
					if ($translation) {
						if ($translation['__metadata']->state == -3) {
							unset($rows[$rowkey]);
							continue 2;
						}
					}
					
					foreach($table['fields'] as $fieldid=>$field) {
						if (isset($field->orgname)) {
							$fieldName = $field->orgname;
						} else {
							$fieldName = $field->name;
						}
						
						// Remove JD_MAGIC_KEY fields used to get primary key in table
						if (strpos($field->name, 'JD_MAGIC_KEY') !== false) {
							unset($row[$fieldid]);
							continue;
						}
						
						if ($translation) {
							if (array_key_exists($fieldName, $translation)) {
								//used for params
								// @todo rework with recrusive version
								if (is_array($translation[$fieldName])) {
									//works only for json atm
									$jsontmp = json_decode($row[$fieldid], true);
									foreach($translation[$fieldName] as $subkey=>$subvalue) {
										$jsontmp[$subkey] = $subvalue;
									}
									$row[$fieldid] = json_encode($jsontmp);
								} else {
									$row[$fieldid] = $translation[$fieldName];
								}
							}
						}
						
						if ($key) {
							if (is_numeric($key)) {
								if ($key==$fieldid) {
									$targetkey = $row[$fieldid];
								}
							} elseif (is_string($key)) {
								if (strtolower($key)==strtolower($fieldName)) {
									$targetkey = $row[$fieldid];
								}
							}
						}
					}
					
					//second turn to copy table to correct key and fieldnames
					foreach($table['fields'] as $fieldid=>$field) {
						// Remove JD_MAGIC_KEY fields used to get primary key in table
						if (strpos($field->name, 'JD_MAGIC_KEY') !== false) {
							continue;
						}
						switch(strtoupper($resulttype)) {
							case 'BOTH':
									$newrows[$targetkey][$fieldid] = $row[$fieldid];
									$newrows[$targetkey][$field->name] = $row[$fieldid];
								break;
							case 'NUM':
									$newrows[$targetkey][$fieldid] = $row[$fieldid];
								break;
							case 'ASSOC':
									$newrows[$targetkey][$field->name] = $row[$fieldid];
								break;
						}
					}
				}
			}
		}
		unset($row);

		if (is_numeric($key) && $key < 0) {
			$rows = current($newrows);
		} else {
			$rows = $newrows;
		}

	}
	
	public function getTranslation($reftable, $refid, $language) {
		if (array_key_exists($reftable, $this->cache)) {
			if (array_key_exists($refid, $this->cache[$reftable])) {
				if (array_key_exists($language, $this->cache[$reftable][$refid])) {
					return $this->cache[$reftable][$refid][$language];
				}
			}
		}
		
		/* caching we may use later if we have a cache cleaning function
		$query = $this->getQuery(true);
		$query->select('distinct referenceTable');
		$query->from('#__jd_store');
		
		$this->setQuery($query);
		$old = $this->setTranslate(false);
		$this->jd_tables = $this->loadResultArray(0);
		$this->setTranslate($old);
		if (!$this->jd_tables) {
			$this->jd_tables = array();
		}
		*/
		
		
		$db = JFactory::getDbo();
		$query = $db->getQuery(true);
		$query->select('*');
		$query->from('#__jd_store');
		$query->where('idLang='.$db->quote($language));
		$query->where('idReference='.$db->quote($refid));
		$query->where('referenceTable='.$db->quote($reftable));
		
		//save current state
		$old = $db->setTranslate(false);
		$oldquery = $db->getQuery();

		$db->setQuery($query);
		$translation = $db->loadObject();

		//restore old state
		$db->setTranslate($old);
		$db->setQuery($oldquery);

		if ($translation) {
			$this->cache[$reftable][$refid][$language] = jDictionTranslationHelper::decodeTranslation($translation->value);
			unset($translation->value);
			$this->cache[$reftable][$refid][$language]['__metadata'] = $translation;
		
			return $this->cache[$reftable][$refid][$language];
		}
		return false;
	}

	/**
	 * returns the status of the translation
	 * @param string table name
	 * @param int table primary key
	 * @param mixed language as lang_code or lang_id or object
	 */
	public function getTranslationStatus($reftable, $refid, $language=null) {
		$status = 0; //
		if ($language) {
			if (!is_object($language)) {
				if (is_numeric($language)) {
					$languages = $this->getLanguages(true);
					foreach($languages as $lang) {
						if ($lang->lang_code == $language) {
							$languages = array($lang);
						}
					}
				} else {
					$lang = new StdClass;
					$lang->lang_id = $language;
					$languages = array($lang);
				}
			} else {
				$languages = array($language);
			}
		} else {
			$languages = $this->getLanguages();
		}
		foreach($languages as $language) {
			$translation = $this->getTranslation($reftable, $refid, $language->lang_id);
			if ($translation) {
				if ($translation->old) {
					$status = JDICTION_TRANSLATION_STATUS_OLD;
				} else {
					$status = JDICTION_TRANSLATION_STATUS_FULL;
				}
			} else {
				$status = JDICTION_TRANSLATION_STATUS_NONE;
			}
		}
		return $status;
	}
	
	public function getLanguages($all=false) {
		$languages = JLanguageHelper::getLanguages();
		if (!$all) {
			$config = JFactory::getConfig();
			foreach($languages as $k=>$language) {
				if ($language->lang_code == $config->get('language')) {
					unset($languages[$k]);
				}
			}
		}
		
		return $languages;
	}
	
	/**
	 * Adds the translation toolbar button to the toolbar based on the
	 * given parameters.
	 *
	 */
	public function addToolbar() {
		//check if we are in backend
		$app = JFactory::getApplication();
		if (!$app->isAdmin()) {
			return;
		}

		$component = JRequest::getCmd('option', false);
		$view 	= JRequest::getCmd('view', false);
		$layout = JRequest::getString('layout', 'default');
		
		$option = $this->getComponent($component);
		if (!$option || !$view || !$layout) {
			return;
		}

		$id = JRequest::getInt($option['__metadata']['key'], 0);

		//Load ToolBar
		$bar = JToolBar::getInstance('toolbar');
		
		//Add Stylesheet for button icons
		JHTML::_('stylesheet','administrator/components/com_jdiction/assets/style.css', array(), false); 
		
		if (array_key_exists($view, $this->components[$component])) {
			if (array_key_exists($layout, $this->components[$component][$view])) {
				if ($this->components[$component][$view][$layout]) {
					if ($id == 0) {
						$bar->appendButton('Link', 'jdiction-component', 'LIB_JDICTION_TOOLBAR_TRANSLATE', "javascript:alert('".JText::_('LIB_JDICTION_SAVE_FIRST', true)."');");
					} else {
						$bar->appendButton('Popup', 'jdiction-component', 'LIB_JDICTION_TOOLBAR_TRANSLATE', 'index.php?option=com_jdiction&view=translation&tmpl=component&layout=popup&jd_option='.$component.'&jd_view='.$view.'&jd_layout='.$layout.'&jd_id='.$id, 'window.getSize().x-100', 'window.getSize().y-100');
					}
				}
			}
		}
	}
	
	public function getTranslationLink($component, $view, $layout, $id, $popup=true) {
		
		$link = 'index.php?option=com_jdiction&view=translation';
		if ($popup) {
			$link .= '&tmpl=component&layout=popup';
		} else {
			$link .= '&layout=edit';
		}
		$link .= '&jd_option='.$component.'&jd_view='.$view.'&jd_layout='.$layout.'&jd_id='.$id;
		
		return $link;
	}
}